# -*- coding: utf-8 -*-
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

# ------------------------------------------------------------------------------
import sys

from appy.px import Px
from appy.fields import Field
from appy.gen.layout import Layouts
from appy.shared.diff import HtmlDiff
from appy.shared.xml_parser import XhtmlCleaner
from appy.fields.multilingual import Multilingual
from appy.gen.indexer import XhtmlTextExtractor, ValidValue

# ------------------------------------------------------------------------------
NOT_IMPLEMENTED = 'This feature is not implemented yet'

# ------------------------------------------------------------------------------
class Rich(Multilingual, Field):
    '''Field allowing to encode a "rich" text, based on XHTML and external
       editor "ckeditor".'''

    # Required Javascript files
    cdnUrl = '//cdn.ckeditor.com/%s/%s/ckeditor.js'

    # Use this constant to say that there is no maximum size for a string field
    NO_MAX = sys.maxint

    # Default ways to render multilingual fields
    defaultLanguagesLayouts = {
      'edit': 'horizontal', 'view': 'horizontal', 'cell': 'vertical'}

    # Override this dict, defined at the Multilingual level
    lgTop = {'view': 1, 'edit': 1}

    # Default styles to use
    defaultStyles = ('p', 'h1', 'h2', 'h3', 'h4')

    # Default custom style, allowing to produce, in POD results, a page break
    # just after the paragraph having this style.
    customStyles = [{'id': 'div', 'element': 'div',
                     'attributes': {'style': 'page-break-after:always'}}]

    # Unilingual pxView and pxCell
    pxViewUni = pxCellUni = Px('''
     <x var="inlineEdit=field.getAttribute(obj, 'inlineEdit');
             mayAjaxEdit=inlineEdit and (layoutType != 'cell') and not \
                         showChanges and zobj.mayEdit(field.writePermission)">
      <div if="not mayAjaxEdit" class="xhtml">::value or '-'</div>
      <x if="mayAjaxEdit" var2="name=lg and ('%s_%s' % (name, lg)) or name">
       <div class="xhtml" contenteditable="true"
            id=":'%s_%s_ck' % (zobj.id, name)">::value or '-'</div>
       <script if="mayAjaxEdit">::field.getJsInlineInit(zobj, name, lg)</script>
      </x>
     </x>''')

    # Unilingual pxEdit
    pxEditUni = Px('''
     <textarea var="inputId=not lg and name or '%s_%s' % (name, lg)"
       id=":inputId" name=":inputId" cols=":field.getTextareaCols()"
       style=":field.getTextareaStyle()"
       rows=":field.height">:field.getInputValue(inRequest, requestValue, value)
     </textarea>
     <script>::field.getJsInit(zobj, lg)</script>''')

    pxSearch = Px('''
     <input type="text" maxlength=":field.maxChars" size=":field.swidth"
            value=":field.sdefault" name=":widgetName"/>''')

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, show=True, page='main', group=None, layouts=None,
      move=0, indexed=False, mustIndex=True, indexValue=None, searchable=False,
      specificReadPermission=False, specificWritePermission=False, width=None,
      height=None, maxChars=None, colspan=1, master=None, masterValue=None,
      focus=False, historized=False, mapping=None, generateLabel=None,
      label=None, sdefault='', scolspan=1, swidth=None, sheight=None,
      persist=True, styles=defaultStyles, customStyles=None,
      allowImageUpload=False, spellcheck=False, languages=('en',),
      languagesLayouts=None, inlineEdit=False, toolbar='Standard', view=None,
      cell=None, edit=None, xml=None, translations=None):
        # The list of styles that the user will be able to select in the styles
        # dropdown (within CKEditor) is defined hereafter.
        self.styles = styles
        # If you want to add custom, non-standard styles in the above-mentioned
        # dropdown, do not touch attribute "styles", but specify such entries in
        # the following attribute "customStyles". It must be a list or dict of
        # entries; every entry represents a custom style and must be a dict
        # having the following keys and values. Every key and value must be a
        # string.
        # ----------------------------------------------------------------------
        #    "id"    | An identifier for your style, that must be different from
        #            | any standard CKEditor style like h1, h2, p, div, etc;
        # ----------------------------------------------------------------------
        #   "name"   | A translated name for your style, that will appear in the
        #            | dropdown. Do not use any special (ie, accentuated) char
        #            | in the value: prefer the use of an HTML entity for
        #            | defining such char.
        # ----------------------------------------------------------------------
        #  "element" | The HTML tag that will surround the text onto which the
        #            | style will be applied. Do not use "blockquote", it does
        #            | not work. Non-standard tag "address" works well, or any
        #            | standard tag like h1, h2, etc. Note that if you use a
        #            | standard tag, it will work, but if you have also
        #            | activated it in attribute "style" hereabove, you won't be
        #            | able to make the difference in the result between both
        #            | styles because they will produce a similar result.
        # ----------------------------------------------------------------------
        self.customStyles = customStyles or Rich.customStyles
        # Do we allow the user to upload images in it ?
        self.allowImageUpload = allowImageUpload
        if allowImageUpload: raise Exception(NOT_IMPLEMENTED)
        # Do we run the CK spellchecker ?
        self.spellcheck = spellcheck
        # What toolbar is used ? Possible values are "Standard" or "Simple"
        self.toolbar = toolbar
        # If "languages" holds more than one language, the field will be
        # multi-lingual and several widgets will allow to edit/visualize the
        # field content in all the supported languages. The field is also used
        # by the CK spell checker.
        self.languages = languages
        # When content exists in several languages, how to render them? Either
        # horizontally (one below the other), or vertically (one besides the
        # other). Specify here a dict whose keys are layouts ("edit", "view")
        # and whose values are either "horizontal" or "vertical".
        self.languagesLayouts = languagesLayouts
        # Call the base constructor
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          searchable, specificReadPermission, specificWritePermission, width,
          height, maxChars, colspan, master, masterValue, focus, historized,
          mapping, generateLabel, label, sdefault, scolspan, swidth, sheight,
          persist, inlineEdit, view, cell, edit, xml, translations)
        # No max chars by default
        if maxChars is None:
            self.maxChars = Rich.NO_MAX

    def getDefaultLayouts(self):
        '''Returns the default layouts for a rich field'''
        # Is this field in a grid-style group ?
        inGrid = self.inGrid()
        if self.historized:
            # self.historized can be a method or a boolean. If it is a method,
            # it means that under some condition, historization will be enabled.
            # So we come here also in this case.
            r = inGrid and 'gc' or 'c'
        else:
            r = inGrid and 'g' or 'b'
        return getattr(Layouts.Rich, r)

    def getDiffValue(self, obj, value, language):
        '''Returns a version of p_value that includes the cumulative diffs
           between successive versions. If the field is non-multilingual, it
           must be called with p_language being None. Else, p_language
           identifies the language-specific part we will work on.'''
        res = None
        lastEvent = None
        name = language and ('%s-%s' % (self.name, language)) or self.name
        for event in obj.workflow_history['appy']:
            if event['action'] != '_datachange_': continue
            if name not in event['changes']: continue
            if res is None:
                # We have found the first version of the field
                res = event['changes'][name][0] or ''
            else:
                # We need to produce the difference between current result and
                # this version.
                iMsg, dMsg = obj.getHistoryTexts(lastEvent)
                thisVersion = event['changes'][name][0] or ''
                comparator = HtmlDiff(res, thisVersion, iMsg, dMsg)
                res = comparator.get()
            lastEvent = event
        if not lastEvent:
            # There is no diff to show for this p_language.
            return value
        # Now we need to compare the result with the current version.
        iMsg, dMsg = obj.getHistoryTexts(lastEvent)
        comparator = HtmlDiff(res, value or '', iMsg, dMsg)
        return comparator.get()

    def getUniFormattedValue(self, obj, value, layoutType='view',
          showChanges=False, userLanguage=None, language=None):
        '''If no p_language is specified, this method is called by
           m_getFormattedValue for getting a non-multilingual value (ie, in
           most cases). Else, this method returns a formatted value for the
           p_language-specific part of a multilingual value.'''
        if Field.isEmptyValue(self, obj, value) and not showChanges: return ''
        r = value
        if showChanges:
            # Compute the successive changes that occurred on p_value
            r = self.getDiffValue(obj, r, language)
        return r

    def extractText(self, value, lower=True, dash=False):
        '''Extracts pure text from XHTML p_value'''
        extractor = XhtmlTextExtractor(lower=lower, dash=dash,
                                       raiseOnError=False)
        return extractor.parse('<p>%s</p>' % value)

    def getUniIndexValue(self, obj, value, forSearch):
        '''Gets the value to index for unilingual value p_value'''
        return self.extractText(value) or ''

    def getUniStorableValue(self, obj, value):
        '''Gets the p_value as can be stored in the database within p_obj'''
        if not value: return value
        # Clean the value. When image upload is allowed, ckeditor inserts some
        # "style" attrs (ie for image size when images are resized). So in this
        # case we can't remove style-related information.
        try:
            value = XhtmlCleaner().clean(value)
        except XhtmlCleaner.Error, e:
            # Errors while parsing p_value can't prevent the user from storing
            # it.
            pass
        # Manage maxChars
        max = self.maxChars
        if max and (len(value) > max): value = value[:max]
        return value

    def getIndexType(self): return 'XhtmlIndex'

    def getTextareaStyle(self):
        '''On "edit", get the content of textarea's "style" attribute'''
        # If the width is expressed as a string, the field width must be
        # expressed in attribute "style" (so returned by this method) and not
        # via attribute "cols" (returned by m_getTextareaCols below).
        return isinstance(self.width, str) and ('width:%s' % self.width) or ''

    def getTextareaCols(self):
        '''When this widget must be rendered as an HTML field of type
           "textarea", get the content of its "cols" attribute.'''
        # Use this attribute only if width is expressed as an integer value
        return isinstance(self.width, int) and self.width or ''

    ckLanguages = {'en': 'en_US', 'pt': 'pt_BR', 'da': 'da_DK', 'nl': 'nl_NL',
                   'fi': 'fi_FI', 'fr': 'fr_FR', 'de': 'de_DE', 'el': 'el_GR',
                   'it': 'it_IT', 'nb': 'nb_NO', 'pt': 'pt_PT', 'es': 'es_ES',
                   'sv': 'sv_SE'}

    def getCkLanguage(self, obj, language):
        '''Gets the language for CK editor SCAYT. p_language is one of
           self.languages if the field is multilingual, None else. If p_language
           is not supported by CK, we use english.'''
        if not language:
            language = self.getAttribute(obj, 'languages')[0]
        if language in self.ckLanguages: return self.ckLanguages[language]
        return 'en_US'

    def getCkParams(self, obj, language):
        '''Gets the base params to set on a rich text field'''
        tool = obj.getTool()
        base = tool.getSiteUrl()
        ckAttrs = {'customConfig': '%s/ui/ckeditor/config.js' % base,
                   'contentsCss': '%s/ui/ckeditor/contents.css' % base,
                   'stylesSet': '%s/ui/ckeditor/styles.js' % base,
                   'toolbar': self.toolbar, 'format_tags':';'.join(self.styles),
                   'scayt_sLang': self.getCkLanguage(obj, language)}
        if self.width: ckAttrs['width'] = self.width
        if self.height: ckAttrs['height'] = self.height
        if self.spellcheck: ckAttrs['scayt_autoStartup'] = True
        # Add custom styles
        if self.customStyles:
            for style in self.customStyles:
                id = style['id']
                ckAttrs['format_%s' % id] = style
            ckAttrs['format_tags'] += ';%s' % id
        if self.allowImageUpload:
            ckAttrs['filebrowserUploadUrl'] = '%s/upload' % obj.absolute_url()
        if not tool.getUser().has_role('Manager'):
            ckAttrs['removeButtons'] = 'Source'
        ck = []
        for k, v in ckAttrs.iteritems():
            if isinstance(v, int): sv = str(v)
            elif isinstance(v, bool): sv = str(v).lower()
            elif isinstance(v, dict): sv = str(v)
            else: sv = '"%s"' % v
            ck.append('%s: %s' % (k, sv))
        return ', '.join(ck)

    def getJs(self, layoutType, r, config):
        if layoutType not in ('edit', 'view'): return
        # Compute the URL to ckeditor CDN
        ckUrl = Rich.cdnUrl % (config.ckVersion, config.ckDistribution)
        if ckUrl not in r: r.append(ckUrl)

    def getJsInit(self, obj, language):
        '''Gets the Javascript init code for displaying ckeditor for this field.
           If the field is multilingual, we must init the rich text editor for a
           given p_language (among self.languages). Else, p_languages is
           None.'''
        name = not language and self.name or ('%s_%s' % (self.name, language))
        return 'CKEDITOR.replace("%s", {%s})' % \
               (name, self.getCkParams(obj, language))

    def getJsInlineInit(self, obj, name, language):
        '''Gets the Javascript init code for enabling inline edition of this
           field If the field is multilingual, the current p_language is given
           and p_name includes it. Else, p_language is None.'''
        id = obj.id
        fieldName = language and name.rsplit('_',1)[0] or name
        lg = language or ''
        return "CKEDITOR.disableAutoInline = true;\n" \
               "CKEDITOR.inline('%s_%s_ck', {%s, on: {blur: " \
               "function( event ) { var content = event.editor.getData(); " \
               "doInlineSave('%s','%s','%s','view',true,content,'%s')}}})" % \
               (id, name, self.getCkParams(obj, language), id, fieldName,
                obj.absolute_url(), lg)
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import difflib, xml.sax
from xml.sax.handler import ContentHandler

from appy.xml import xmlPrologue

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
DIFFS_DETECTED = 'Difference(s) detected between files %s and %s:'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class XmlHandler(ContentHandler):
    '''Handler is used for producing, in self.r, a readable XML (with carriage
       returns) and for removing some tags that always change (like dates) from
       a file that need to be compared to another file.'''

    def __init__(self, xmlTagsToIgnore, xmlAttrsToIgnore):
        ContentHandler.__init__(self)
        self.r = xmlPrologue
        self.namespaces = {} # ~{s_namespaceUri: s_namespaceName}~
        self.indentLevel = -1
        self.tabWidth = 3
        self.tagsToIgnore = xmlTagsToIgnore
        self.attrsToIgnore = xmlAttrsToIgnore
        # Some content must be ignored, and not dumped into the result
        self.ignoring = False

    def isIgnorable(self, tag):
        '''Is p_tag an ignorable tag ?'''
        r = False
        for tagName in self.tagsToIgnore:
            if isinstance(tagName, list) or isinstance(tagName, tuple):
                # We have a namespace
                nsUri, name = tagName
                try:
                    nsName = self.ns(nsUri)
                    tagFullName = '%s:%s' % (nsName, name)
                except KeyError:
                    tagFullName = ''
            else:
                # No namespace
                tagFullName = tagName
            if tagFullName == tag:
                r = True
                break
        return r

    def setDocumentLocator(self, locator):
        self.locator = locator

    def endDocument(self): pass

    def dumpSpaces(self):
        self.r += '\n' + (' ' * self.indentLevel * self.tabWidth)

    def manageNamespaces(self, attrs):
        '''Manage namespaces definitions encountered in p_attrs'''
        for name, value in attrs.items():
            if name.startswith('xmlns:'):
                self.namespaces[value] = name[6:]

    def ns(self, nsUri):
        return self.namespaces[nsUri]

    def startElement(self, tag, attrs):
        self.manageNamespaces(attrs)
        # Do we enter into a ignorable element ?
        if self.isIgnorable(tag):
            self.ignoring = True
        else:
            if not self.ignoring:
                self.indentLevel += 1
                self.dumpSpaces()
                self.r += '<%s' % tag
                attrsNames = attrs.keys()
                attrsNames.sort()
                for attrToIgnore in self.attrsToIgnore:
                    if attrToIgnore in attrsNames:
                        attrsNames.remove(attrToIgnore)
                for attrName in attrsNames:
                    self.r += ' %s="%s"' % (attrName, attrs[attrName])
                self.r += '>'

    def endElement(self, tag):
        if self.isIgnorable(tag):
            self.ignoring = False
        else:
            if not self.ignoring:
                self.dumpSpaces()
                self.indentLevel -= 1
                self.r += '</%s>' % tag

    def characters(self, content):
        if not self.ignoring:
            self.r += content.replace('\n', '')

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Diff:
    '''Compares 2 XML files and produces a diff'''

    def __init__(self, fileNameA, fileNameB, areXml=True,
                 xmlTagsToIgnore=(), xmlAttrsToIgnore=()):
        self.fileNameA = fileNameA
        self.fileNameB = fileNameB
        self.areXml = areXml # Can also diff non-XML files
        self.xmlTagsToIgnore = xmlTagsToIgnore
        self.xmlAttrsToIgnore = xmlAttrsToIgnore

    def say(self, message, report=None):
        '''Outputs a message, either into the p_report or on stdout'''
        if report:
            report.say(message)
        else:
            print(message)

    def filesAreIdentical(self, report=None):
        '''Compares the 2 files and returns True if they are identical (if we
           ignore xmlTagsToIgnore and xmlAttrsToIgnore). If p_report is
           specified, it must be an instance of appy.test.report.TestReport;
           the diffs will be dumped in it.'''
        # Perform the comparison
        differ = difflib.Differ()
        if self.areXml:
            f = open(self.fileNameA)
            contentA = f.read()
            f.close()
            f = open(self.fileNameB)
            contentB = f.read()
            f.close()
            xmlHandler = XmlHandler(self.xmlTagsToIgnore, self.xmlAttrsToIgnore)
            xml.sax.parseString(contentA, xmlHandler)
            contentA = xmlHandler.r.split('\n')
            xmlHandler = XmlHandler(self.xmlTagsToIgnore, self.xmlAttrsToIgnore)
            xml.sax.parseString(contentB, xmlHandler)
            contentB = xmlHandler.r.split('\n')
        else:
            f = open(self.fileNameA)
            contentA = f.readlines()
            f.close()
            f = open(self.fileNameB)
            contentB = f.readlines()
            f.close()
        diffResult = list(differ.compare(contentA, contentB))
        # Analyse, format and report the result
        atLeastOneDiff = False
        lastLinePrinted = False
        i = -1
        for line in diffResult:
            i += 1
            if line and (line[0] != ' '):
                if not atLeastOneDiff:
                    msg = DIFFS_DETECTED % (self.fileNameA, self.fileNameB)
                    self.say(msg, report)
                    atLeastOneDiff = True
                if not lastLinePrinted:
                    self.say('...', report)
                if self.areXml:
                    self.say(line, report)
                else:
                    self.say(line[:-1], report)
                lastLinePrinted = True
            else:
                lastLinePrinted = False
        return not atLeastOneDiff
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Functions for sending emails'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import smtplib, socket, time
from email import encoders
from email.header import Header
from email.utils import formatdate
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
NO_CONFIG  = 'Must send mail but no SMTP server configured.'
NO_SERVER  = 'no mailhost defined'
DISABLED   = 'Mail disabled%s: should send mail from %s to %d recipient(s): %s.'
REPLY_TO   = 'reply to: %s'
MSG_SUBJ   = 'Subject: %s'
MSG_BODY   = 'Body: %s'
MSG_ATTS   = '%d attachment(s).'
MSG_SEND   = 'Sending mail from %s to %s (subject: %s).'
MAIL_R_KO  = 'Could not send mail to some recipients. %s'
MAIL_SENT  = "Mail sent in %.2f''."
MAIL_NSENT = '%s: mail sending failed (%s).'
CONNECT_OK = "Connected to %s in %.2f''."

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Config:
    '''Parameters for connecting to a SMTP server'''

    # Currently we don't check the connection to the SMTP server at startup
    testable = False

    def __init__(self, fromName=None, fromEmail='info@appyframework.org',
                 replyTo=None, server='localhost', port=25, secure=False,
                 login=None, password=None, enabled=True):
        # The name that will appear in the "from" part of the messages
        self.fromName = fromName
        # The mail address that will appear in the "from" part of the messages
        self.fromEmail = fromEmail
        # The optional "reply-to" mail address
        self.replyTo = replyTo
        # The SMTP server address and port
        if ':' in server:
            self.server, port = server.split(':')
            self.port = int(port)
        else:
            self.server = server
            self.port = int(port) # That way, people can specify an int or str
        # Secure connection to the SMTP server ?
        self.secure = secure
        # Optional credentials to the SMTP server
        self.login = login
        self.password = password
        # Is this server connection enabled ?
        self.enabled = enabled

    def init(self, tool): pass

    def getFrom(self):
        '''Gets the "from" part of the messages to send.'''
        if self.fromName: return '%s <%s>' % (self.fromName, self.fromEmail)
        return self.fromEmail

    def __repr__(self):
        '''Short string representation of this mail config, for logging and
           debugging purposes.'''
        r = 'mail: %s:%d' % (self.server, self.port)
        if self.login: r += ' (login as %s)' % self.login
        return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def logMailSent(r, start, log):
    '''A mail a just been sent. Log its success or failure (depending on p_r)'''
    if not log: return
    if r:
        log(MAIL_R_KO % str(r), type='warning')
    else:
        log(MAIL_SENT % (time.time() - start))

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def sendMail(config, to, subject, body, attachments=None, log=None,
             replyTo=None, split=False):
    '''Sends a mail, via the SMTP server defined in the p_config'''

    # p_config is an instance of class appy.utils.mail.Config hereabove.

    # This function sends a mail to p_to (a single email recipient or a list of
    # recipients). Every (string) recipient can be an email address or a string
    # of the form "[name] <[email]>".

    # p_attachments must be a list or tuple whose elements can have 2 forms:
    # 1. a tuple (fileName, fileContent): "fileName" is the name of the file
    #    as a string; "fileContent" is the file content, also as a string;
    # 2. an instance of class appy.model.fields.file.FileInfo.

    # p_log can be a function accepting 2 args:
    # - the message to log (as a string);
    # - the second must be named "type" and will receive string
    #   "info", "warning" or "error".

    # A p_replyTo mail address or recipient can be specified

    # If p_split is False, a single mail will be send to all repicients
    # specified at once in field "To". Else, a distinct mail will be sent to
    # every recipient defined in p_to.

    to = [to] if isinstance(to, str) else to
    if not config:
        if log: log(NO_CONFIG)
        return
    # Just log things if mail is disabled
    fromAddress = config.getFrom()
    replyTo = replyTo or config.replyTo
    if not config.enabled or not config.server:
        msg = '' if config.server else ' (%s)' % NO_SERVER
        if log:
            toLog = DISABLED % (msg, fromAddress, len(to), ', '.join(to))
            if replyTo: toLog += ' (%s).' % (REPLY_TO % replyTo)
            log(toLog)
            log(MSG_SUBJ % subject)
            log(MSG_BODY % body)
        if attachments and log: log(MSG_ATTS % len(attachments))
        return
    if log: log(MSG_SEND % (fromAddress, ', '.join(to), subject))
    # Create the base MIME message
    body = MIMEText(body, 'plain')
    if attachments:
        msg = MIMEMultipart()
        msg.attach(body)
    else:
        msg = body
    # Add the header values
    msg['Subject'] = Header(subject)
    msg['From'] = fromAddress
    msg['Date'] = formatdate(localtime=True)
    if replyTo: msg.add_header('reply-to', replyTo)
    # Add attachments
    if attachments:
        for attachment in attachments:
            # 2 possible forms for an attachment
            if isinstance(attachment, tuple) or isinstance(attachment, list):
                fileName, fileContent = attachment
            else:
                # a FileInfo instance
                fileName = attachment.uploadName
                f = file(attachment.fsPath, 'rb')
                fileContent = f.read()
                f.close()
            part = MIMEBase('application', 'octet-stream')
            part.set_payload(fileContent)
            encoders.encode_base64(part)
            part.add_header('Content-Disposition',
                            'attachment; filename="%s"' % fileName)
            msg.attach(part)
    # Send the mail(s)
    try:
        start = time.time()
        # Connect to the SMTP server
        smtpServer = smtplib.SMTP(config.server, port=config.port)
        if config.secure:
            smtpServer.ehlo()
            smtpServer.starttls()
        if config.login:
            smtpServer.login(config.login, config.password)
        # Log the time spent while connecting, if we perform a single connection
        # for sending several mails.
        if split and log:
            log(CONNECT_OK % (time.time()-start))
        # Precompute the mail body if a single mail must be sent
        if not split:
            msg['To'] = ', '.join(to)
            body = msg.as_string()
        # Send the mail(s)
        if split:
            for recipient in to:
                start = time.time()
                msg['To'] = recipient
                body = msg.as_string()
                r = smtpServer.sendmail(fromAddress, [recipient], body)
                logMailSent(r, start, log)
        else:
            r = smtpServer.sendmail(fromAddress, to, body)
            logMailSent(r, start, log)
        # Disconnect from the server
        smtpServer.quit()
    except smtplib.SMTPException as e:
        if log:
            log(MAIL_NSENT % (config, str(e)), type='error')
    except socket.error as se:
        if log:
            log(MAIL_NSENT % (config, str(se)), type='error')
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#!/usr/bin/python3

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from pathlib import Path
from appy.bin import Program
from appy.deploy import Deployer

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Deploy(Program):
    '''This program allows to deploy apps in sites on distant servers'''

    # Available commands
    COMMANDS = {'list':    'Lists the available targets.',
                'info':    'Retrieve info about the target OS',
                'install': 'Install required dependencies via the OS ' \
                           'package system (currently: "apt" only).',
                'site':    'Create an Appy site on the target.',
                'update':  'Updates a site and (re)start it.',
                'view':    "View app.log's tail on the target"
    }

    def helpCommands(COMMANDS):
        '''Builds the text describing available commands'''
        r = []
        for command, text in COMMANDS.items():
            r.append('"%s": %s' % (command, text))
        return '\n'.join(r)

    # Help messages
    HELP_APP     = 'The path, on this machine, to the app to deploy ' \
                   '(automatically set if called from <site>/bin/deploy).'
    HELP_SITE    = 'The path, on this machine, to the reference site ' \
                   '(automatically set if called from <site>/bin/deploy).'
    HELP_COMMAND = 'The command to perform. Available commands are: %s' % \
                   helpCommands(COMMANDS)
    HELP_TARGET  = 'The target to deploy to. By default, target "default" ' \
                   'will be chosen.'

    # Error messages
    FOLDER_KO    = '%s does not exist or is not a folder.'
    COMMAND_KO   = 'Command "%s" does not exist.'

    def defineArguments(self):
        '''Define the allowed arguments for this program'''
        parser = self.parser
        # Positional arguments
        parser.add_argument('app', help=Deploy.HELP_APP)
        parser.add_argument('site', help=Deploy.HELP_SITE)
        parser.add_argument('command', help=Deploy.HELP_COMMAND)
        # Optional arguments
        parser.add_argument('-t', '--target', dest='target',
                            help=Deploy.HELP_TARGET)

    def analyseArguments(self):
        '''Check and store arguments'''
        args = self.args
        # Check and get the paths to the app and site
        for name in ('app', 'site'):
            path = Path(getattr(args, name))
            if not path.is_dir():
                self.exit(self.FOLDER_KO % path)
            setattr(self, name, path)
        self.command = args.command
        if self.command not in Deploy.COMMANDS:
            self.exit(self.COMMAND_KO % self.command)
        self.target = args.target

    def run(self):
        return Deployer(self.app, self.site, self.command, self.target).run()

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
if __name__ == '__main__': Deploy().run()
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Operators for object searches'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from DateTime import DateTime
from BTrees.IIBTree import multiunion, intersection

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Searching objects in the database is done via method
# appy.model.base.Base::search and variants. This method accepts keyword
# arguments to define such searches. These arguments mention indexed fields.
# Workflow state is one example of default indexed field. For example, searching
# all users being active can be done via any object, like this:
#
#               activeUsers = o.search('User', state='active')
#
# On this example, keyword argument named "state" takes the value "active".
# Instead of specifying such a single value, one may use operators to define a
# set or range of values. Such operators are defined in this module. For
# example, the "or" operator could be used as value for keyword argument "state"
# to retrieve users being in states "active" or "inactive", like this:
#
#         moreUsers = o.search('User', state=or_('active', 'inactive'))
#
# As PEP8 recommends it, the operator, also being a Python reserved keyword, is
# suffixed with an underscore. The search hereabove retrieves all users being
# active or inactive.
#
# For multivalued fields, operator "and" can also be used. In the following
# example, we suppose the existence of an indexed field named "attributes" on
# class User, that can hold one or more values among
#
#                      ["married", "employed", "faithful"]
# 
# Getting all users being married as well as employed, being active or not, is
# done via:
#
# someUsers = o.search('User', state=or_('active', 'inactive'),
#                              attributes=and_('married', 'employed'))
#
# The "in" operator allow to defined ranges. Getting all users having an IQ
# between 130 and 155 is expressed like this:
#
#                 smartUsers = o.search('User', iq=in_(130, 155))
#
# The "not" operator allows to define all users not statisfying some condition
# based on some indexed field. Suppose the user workflow defines these states:
# active, inactive and registered (=requires validation from a Manager).
# Getting all users not being inactive can be done via:
#
#           awareUsers = o.search('User', state=not_('inactive'))
# 
# The "search" method can accept any number of keyword arguments; values for
# these latters can be simple values or operator values.
#
# Every such operator corresponds to a class in this module. All these classes
# inherit from an abstract base class named Operator. Using operators like in
# the previous examples creates instances of the concrete classes from this
# module. These classes are made available in the appy.all module, so are
# available in any module from your app where you have used the standard import
# statement:
#                           from appy.all import *

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
RANGE_KO = 'Exactly 2 values must be provided: the lower and upper bounds.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Operator:
    '''Abstract base class for any operator'''

    # An operator being "negative" produces a set of objects that must be
    # excluded from the result.
    negative = False

    def __init__(self, *values):
        # The list of p_values onto which to operator is applied
        self.values = values

    def apply(self, index, field, rs):
        '''Apply the operator on p_index and return the result. The global
           result set computed so far can be available in p_rs.'''
        # More precisely, the method must return a tuple (result, rsUpdated):
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #  result   | is the set of matching object IDs after having applied the
        #           | operator instance ;
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # rsUpdated | is True if the operator has already updated the global
        #           | resultset p_rs. It can be the case for some optimizations.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def __repr__(self):
        '''Returns p_self's string representation'''
        values = []
        for val in self.values:
            if isinstance(val, DateTime):
                values.append("DateTime('%s')" % str(val))
            elif isinstance(val, str):
                values.append("'%s'" % val.replace("'", "\\'"))
            else:
                values.append(str(val))
        # By evaluating this string representation, the instance can be reified
        return '%s(%s)' % (self.__class__.__name__, ', '.join(values))

    @classmethod
    def fromRequest(class_, req, name):
        '''Return the operator to use for a Field named p_name whose value has
           been encoded in a search form and is present in p_req.'''
        operKey = 'o_%s' % name
        return and_ if req and (req[operKey] == 'and') else or_

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class or_(Operator):

    # Result sets having a size lower than this size are considered "small"
    RS_LIMIT = 200

    def apply(self, index, field, rs):
        '''Apply this "or" operator instance on values stored in p_index'''
        r = []
        for value in self.values:
            ids = index.getByValue(value, field)
            if ids:
                r.append(ids)
        # If the global resultset p_rs is small, starting with intersecting
        # every set of ids with it and doing the union later is faster than
        # creating a multiunion first.
        if (rs is not None) and (len(rs) < or_.RS_LIMIT):
            sets = []
            for ids in r:
                sets.append(intersection(rs, ids))
            r = multiunion(sets)
            rsUpdated = True
        else:
            r = multiunion(r)
            rsUpdated = False
        return r, rsUpdated

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class and_(Operator):
    def apply(self, index, field, rs):
        '''Apply this "and" operator instance on values stored in p_index'''
        r = []
        for value in self.values:
            ids = index.getByValue(value, field)
            if ids is None:
                # There is no match at all
                return None, True
            r.append(ids)
        # Set smaller sets first
        if len(r) > 2:
            r.sort(key=len)
        # Intersect every set of ids with the global resultset p_rs
        for ids in r:
            rs = intersection(rs, ids)
            if not rs:
                return None, True
        return rs, True

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class in_(Operator):
    '''Defines a range of values by specifying lower and upper bounds'''
    # ~~~
    # If you want to specify a range whose upper bound is defined, but with no
    # lower bound, use value None as lower bound, as in: qi=in_(None, 155).
    # ~~~
    # If you want to specify a range whose lower bound is defined, but with no
    # upper bound, use value None as upper bound, as in: qi=in_(100, None).
    # ~~~

    def __init__(self, *values):
        if len(values) != 2:
            raise Exception(RANGE_KO)
        Operator.__init__(self, *values)

    def apply(self, index, field, rs):
        '''Apply this "in" operator instance on values stored in p_index'''
        # Unwrap the bounds from p_self.values
        lo, hi = self.values
        sets = index.getByRange(lo, hi, field)
        r = sets[0] if len(sets) == 1 else multiunion(sets)
        return r, False

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class not_(Operator):

    # The "not" operator is "negative"
    negative = True

    def apply(self, index, field, rs):
        '''Apply this "not" operator instance on values stored in p_index, in
           the same way as a "or". Afterwards, the result of this "not" operator
           will be substracted from the global resultset.'''
        r = []
        for value in self.values:
            ids = index.getByValue(value, field)
            if ids:
                r.append(ids)
        return multiunion(r), False
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# ------------------------------------------------------------------------------
import sys, os, os.path
import Zope2.Startup.zopectl as zctl

# ------------------------------------------------------------------------------
class ZopeRunner:
    '''This class allows to run a Appy/Zope instance.'''

    def run(self):
        # Check that an arg has been given (start, stop, fg, run)
        if not sys.argv[3].strip():
            print('Argument required.')
            sys.exit(-1)
        # Identify the name of the application for which Zope must run.
        app = os.path.splitext(os.path.basename(sys.argv[2]))[0].lower()
        # Launch Zope.
        options = zctl.ZopeCtlOptions()
        options.realize(None)
        options.program = ['/usr/bin/%srun' % app]
        options.sockname = '/var/lib/%s/zopectlsock' % app
        c = zctl.ZopeCmd(options)
        c.onecmd(" ".join(options.args))
        return min(c._exitstatus, 1)
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.ui import Language
from appy.ui.layout import ColumnLayout
from appy.model.utils import Object as O

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class ColSet:
    '''Represents a named set of columns to show when displaying Search results
       or tied objects from a Ref.'''
    FIELD_NOT_FOUND = 'field "%s", used in a column specifier, not found.'

    # Standard column specifiers
    standardColSpecs = O(
      number =     O(special=True, field='_number', width='15px',
                     align='left', header=False),
      checkboxes = O(special=True, field='_checkboxes', width='10px',
                     align='center', header=True)
    )

    @staticmethod
    def getSpecs(class_, o, columnLayouts, dir, addNumber=False,
                 addCheckboxes=False):
        '''Extracts and returns, from a list of p_columnLayouts, info required
           for displaying columns of field values for instances of p_class_,
           either in a result screen or for a Ref field.
        '''
        # p_columnLayouts are specified for each field whose values must be
        # shown. 2 more, not-field-related, column layouts can be specified with
        # these names:
        # ----------------------------------------------------------------------
        # "_number"     | if the listed objects must be numbered by Appy, this
        #               | string represents the column containing that number;
        # ----------------------------------------------------------------------
        # "_checkboxes" | if Appy must show checkboxes for the listed objects,
        #               | this string represents the column containing the
        #               | checkboxes.
        # ----------------------------------------------------------------------
        # If columns "_number" and "_checkboxes" are not among p_columnLayouts
        # but are required (by p_addNumber and p_addCheckboxes), they are added
        # to the result. Specifying them within p_columnLayouts allows to give
        # them a precise position among all columns. When automatically added,
        # they will appear before any other column (which is desirable in most
        # cases).
        r = []
        numberFound = checkboxesFound = False
        for info in columnLayouts:
            name, width, align, header = ColumnLayout(info).get()
            # It that a special column name ?
            special = True
            if name == '_number': numberFound = True
            elif name == '_checkboxes': checkboxesFound = True
            else: special = False
            align = Language.flip(align, dir)
            # For non-special columns, get the corresponding field
            if not special:
                field = class_.fields.get(name)
                if not field:
                    o.log(ColSet.FIELD_NOT_FOUND % name, type='warning')
                    continue
            else:
                # Let the column name in attribute "field"
                field = name
            r.append(O(special=special, field=field, width=width,
                       align=align, header=header))
        # Add special columns if required and not present
        if addCheckboxes and not checkboxesFound:
            r.insert(0, ColSet.standardColSpecs.checkboxes)
        if addNumber and not numberFound:
            r.insert(0, ColSet.standardColSpecs.number)
        return r

    def __init__(self, identifier, label, columns, specs=False):
        # A short identifier for the set
        self.identifier = identifier
        # The i18n label to use for giving a human-readable name to the set
        self.label = label
        # The list/tuple of columns, expressed as strings. Every string must
        # contain a field name, but can be completed (after a char *) by column
        # width and alignment, as in "state*100px|". The "width" part, just
        # after the *, can hold anything that can be set in a "width" HTML
        # attribute. The last char represents the alignment:
        #   ";"   left-aligned (the default);
        #   "|"   centered;
        #   "!"   right-aligned.
        if not specs:
            self.columns = columns
        else:
            # "specs" is the internal representation of "columns". Do not
            # specify "specs=True". It will contain a list of Object instances
            # instead of strings. Every such instance has splitted string info
            # into fields "field", "width" and "align".
            self.specs = columns
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''By importing this module (from appy.all import *) the Appy developer has all
   base stuff for building his app.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy.model.workflow import *
from appy.utils import No, breakpoint
from appy.model.fields.pod import Pod
from appy.model.searches import Search
from appy.model.fields.info import Info
from appy.model.fields.date import Date
from appy.model.fields.hour import Hour
from appy.model.fields.file import File
from appy.model.fields.list import List
from appy.model.fields.dict import Dict
from appy.model.fields.text import Text
from appy.model.fields.rich import Rich
from appy.model.fields.phase import Page
from appy.model.utils import Object as O
from appy.model.fields.float import Float
from appy.model.fields.color import Color
from appy.model.fields import Field, Show
from appy.ui.layout import Layout, Layouts
from appy.model.fields.string import String
from appy.model.fields.action import Action
from appy.model.fields.switch import Switch
from appy.model.fields.custom import Custom
from appy.model.fields.colset import ColSet
from appy.test.monitoring import Monitoring
from appy.model.workflow.state import State
from appy.model.fields.boolean import Boolean
from appy.model.fields.integer import Integer
from appy.model.fields.ref import Ref, autoref
from appy.model.fields.calendar import Calendar
from appy.model.fields.computed import Computed
from appy.model.fields.password import Password
from appy.model.searches.gridder import Gridder
from appy.model.fields.group import Group, Column
from appy.model.workflow.transition import Transition
from appy.model.fields.select import Select, Selection
from appy.database.operators import or_, and_, in_, not_
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''This module contains command-line programs'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import sys, argparse

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Program:
    '''Abstract base class for all command-line programs'''

    def __init__(self):
        # Create the argument parser
        self.parser = argparse.ArgumentParser()
        # Define arguments
        self.defineArguments()
        # Parse arguments
        self.args = self.parser.parse_args()
        # Analyse parsed arguments
        self.analyseArguments()

    def defineArguments(self):
        '''Override this method to define arguments to self.parser (using
           method self.parser.add_argument): see documentation for the
           argparse module.'''

    def analyseArguments(self):
        '''Override this method to analyse arguments that were parsed an stored
           in self.args:
           * check their validity;
           * potentially apply some computation on it and store, on p_self, new
             attributes derived from self.args.
           If this method detects an error within self.args, abort program
           execution by calling m_exit (see below).
        '''

    def exit(self, msg, printUsage=True):
        '''Exists the program afer an error has been encountered and display
           some p_msg.'''
        print(msg)
        if printUsage:
            self.parser.print_usage()
        return sys.exit(1)

    def run(self):
        '''Executes the program after arguments were successfully analysed'''
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import collections
from appy.px import Px
from appy.utils import iconParts
from appy.model.utils import Object

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Page:
    '''Used for describing a page, its related phase, show condition, etc.'''
    subElements = ('save', 'cancel', 'edit')

    def __init__(self, name, phase='main', show=True, showSave=True,
                 showCancel=True, showEdit=True, label=None, icon=None,
                 sticky='edit'):
        # A short, camel-cased unique page, that must be unique across all
        # phases within the same class
        self.name = name
        # The name of the phase containing this page
        self.phase = phase
        # When must the page be shown ? The following attribute can hold values
        # True, False, "view" or "edit", or a method producing such a value,
        # accepting the current object, as unique arg.
        self.show = show
        # When editing the page, must the "save" button be shown ?
        self.showSave = showSave
        # When editing the page, must the "cancel" button be shown ?
        self.showCancel = showCancel
        # When viewing the page, must the "edit" icon be shown ?
        self.showEdit = showEdit
        # The i18n label may be forced here instead of being deduced from the
        # page name.
        self.label = label
        # The icon to display. If you want to use:
        # * one of the standard Appy icons, residing in appy/ui/static, use
        #   string "appy/<icon_name>", for example "appy/help.svg";
        # * an icon being in your app's "static" folder, use string
        #   "<app_name>/<icon_name>";
        # * an icon that is chosen dynamically, set a method from the class
        #   where this page is defined. It will be called without arg and must
        #   return the icon URL, or None if the icon is not available (yet).
        #   Here are some tips.
        #   1) If you want to build the URL to a standard Appy icon, use, ie:
        #      o.buildUrl('help.svg')
        #      ("o" being any instance of any Appy class)
        #   2) If you want to build the URL to an icon being in your app's
        #      "static" folder, use, ie:
        #      o.buildUrl('myIcon.png', base='MyApp')
        #   3) If you want to build the URL to an icon that was uploaded in a
        #      File field, use:
        #      '%s/<field_name>/download' % o.url
        self.setIcon(icon or 'page.svg')
        # Stickyness refers to the complete block of controls that is shown for
        # the current object, when this page is shown. Attribute "sticky" may
        # hold True, False, "edit" or "view", and determines if the block must
        # be sticky when this page is shown on "edit", "view" or both layouts.
        # When sticky, the block remains visible, even when scrollling down the
        # page. It is enabled by default on "edit" layouts only, because
        # stickyness of the block of controls may be in conflict with stickyness
        # of parts of fields shown on the the same page, ie, headers of tables
        # for Ref or List fields. If several sticky elements are present, when
        # scrolling, they will unsightly overlap.
        self.sticky = sticky

    def clone(self):
        '''Create a clone of p_self'''
        r = Page(self.name)
        for k, v in self.__dict__.items(): setattr(r, k, v)
        return r

    def setIcon(self, icon):
        '''Sets this p_icon for p_self'''
        if callable(icon):
            self.icon = icon
            self.iconBase = None
        else:
            self.icon, self.iconBase = iconParts(icon)

    @staticmethod
    def get(pageData):
        '''Produces a Page instance from p_pageData. User-defined p_pageData
           can be:
           (a) a string containing the name of the page;
           (b) a string containing <pageName>_<phaseName>;
           (c) a Page instance.
           This method returns always a Page instance.'''
        r = pageData
        if r and isinstance(r, str):
            # Page data is given as a string
            pageElems = pageData.rsplit('_', 1)
            if len(pageElems) == 1: # We have case (a)
                r = Page(pageData)
            else: # We have case (b)
                r = Page(pageData[0], phase=pageData[1])
        return r

    def isShowable(self, o, layout, elem='page'):
        '''Is this page showable for p_o on p_layout ("view" or "edit")?

           If p_elem is not "page", this method returns the fact that a
           sub-element is viewable or not (buttons "save", "cancel", etc).'''
        # Define what attribute to test for "showability"
        attr = 'show' if elem == 'page' else 'show%s' % elem.capitalize()
        # Get the value of the "show" attribute as identified above
        r = getattr(self, attr)
        if callable(r):
            r = o.H().methods.call(o, r)
        if isinstance(r, str): return r == layout
        return r

    def nextName(self):
        '''If this page has a name being part of a series (ie numeric:
           1, 2, 3... or alphabetical like a, b, c...), this method returns the
           next name in the series.'''
        name = self.name
        if name.isdigit():
            # Increment this number by one and get the result as a string
            r = str(int(name) + 1)
        elif (len(name) == 1) and name.islower() and (name != 'z'):
            r = chr(ord(name) + 1)
        else:
            # Repeat the same name
            r = name
        return r

    def __repr__(self): return '<Page %s - phase %s>' % (self.name, self.phase)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class UiPage:
    '''A page object built at run time, containing only information tied to the
       execution context.'''

    def __init__(self, uiPhase, page, showOnView, showOnEdit):
        # The link to the container phase
        self.uiPhase = uiPhase
        # The corresponding Page instance
        self.page = page
        self.name = page.name
        # Must the page be shown on view/edit layouts ?
        self.showOnView = showOnView
        self.showOnEdit = showOnEdit
        # Must sub-elements (buttons "save", "cancel", etc) be shown ?
        phases = uiPhase.container
        o = phases.o
        layout = phases.layout
        for elem in Page.subElements:
            showable = page.isShowable(o, layout, elem)
            setattr(self, 'show%s' % elem.capitalize(), showable)
        # Define the icon to display
        self.iconUrl = self.getIconUrl(o)

    def __repr__(self):
        '''p_self's string representation'''
        return '<UiPage %s>' % self.name

    def getLabel(self):
        '''Returns the translated label for this page, potentially based on a
           fixed label if p_self.label is not empty.'''
        page = self.page
        o = self.uiPhase.container.o
        label = page.label or '%s_page_%s' % (o.class_.name, page.name)
        return o.translate(label)

    def showable(self, layout):
        '''Return True if this page is showable on p_layout'''
        return self.showOnEdit if layout == 'edit' else self.showOnView

    def getIconUrl(self, o):
        '''Compute the URL to the icon to display for this page'''
        page = self.page
        # The icon to show
        if callable(page.icon):
            url = page.icon(o)
            if url: return url
            # The icon may not be available yet. For example, we may be creating
            # an object having a field that will be used to store the page icon,
            # but no image has been uploaded yet.
            base = None
            icon = 'page.svg'
        else:
            base, icon = page.iconBase, page.icon
        return o.buildUrl(icon, base)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Phase:
    '''A phase is a group of pages within a class'''

    # This class is used only in order to generate phase- and page-related i18n
    # labels, and as general information about all phases and pages defined for
    # a class in the app's model.

    # It is not used by the Appy developer: if a phase must be defined by him in
    # an app, he will do it via a phase name specified as a string, in attribute
    # "phase" of class Page hereabove.

    # It is not used by Appy at run-time neither: class UiPhases is used
    # instead. Indeed, at run-time, the set of phases and pages which are
    # visible by the logged user can be a sub-set of all statically-defined
    # phases and pages.

    def __init__(self, name):
        self.name = name
        # A phase is made of pages ~{p_name: appy.model.fields.page.Page}~
        self.pages = collections.OrderedDict()

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class UiPhase:
    '''A phase object built at run time, containing only the visible pages
       depending on user's permissions.'''

    view = Px('''
     <div class=":'phase phaseC' if compact else 'phase'"
          if="phase.showPages(singlePhase, isEdit)">

      <!-- Siblings navigation -->
      <div var="nav=req.nav"
           if="not isEdit and (phase.name=='main') and (nav) and (nav != 'no')"
           var2="self=ui.Siblings.get(nav, tool, _ctx_)">:self.pxNavigate</div>

      <!-- Page(s) within the phase -->
      <div if="not isEdit" for="pag in phase.pages.values()"
           var2="isCurrent=pag.name == page"
           id=":'currentPage' if isCurrent else ''"
           class=":'currentPage' if isCurrent else ''">

       <!-- This additional div is used to vertically align its content -->
       <div var="viewUrl=o.getUrl(sub='view',page=pag.name,popup=popup,nav=nav)\
                  if pag.showOnView else None;
                 pictoUrl=pag.iconUrl">

        <!-- Main icon -->
        <div>
         <a if="viewUrl" href=":viewUrl">
          <img src=":pictoUrl" class=":picto"/>
         </a>
         <img if="not viewUrl" src=":pictoUrl" class=":picto"/>
        </div>

        <!-- Page name and icons -->
        <x var="label=Px.truncateText(pag.getLabel(), 25)">
         <a if="viewUrl" href=":viewUrl">::label</a>
         <x if="not viewUrl">::label</x>
        </x>

        <!-- The edit or lock icon -->
        <x var="locked=o.Lock.isSet(o, user, pag.name);
               editable=mayEdit and mayAct and pag.showOnEdit and pag.showEdit">
         <a if="editable and not locked"
            href=":o.getUrl(sub='edit', page=pag.name, popup=popup, nav=nav)">
          <img src=":url('editPage.svg')" class="pictoT shake"
               title=":_('object_edit')"/></a>
         <x var="page=pag.name; lockStyle='iconM pictoT'">::o.Lock.px</x>
        </x>
       </div>
      </div>

      <!-- Additional object controls -->
      <x>:o.pxButtons</x>
     </div>''',

     # Class names ending with "C" are used when controls are *C*ompact
     css='''
      .phase { margin:0 0 30px -30px; display:flex; flex-wrap:wrap;
               width:fit-content }
      .phase div, .phase form { display:|pageDisplay| }

      .phase > div { height:|pageHeight|; width:|pageWidth|; text-align:center;
        text-transform:uppercase; font-weight:bold; font-size:90%;
        display:flex !important; align-items:center; justify-content:center;
        padding:|pagePadding|; background-color:|phaseBgColor|;
        position:relative; padding:0 5px; color:|phaseColor|;
        border-right:1px solid |phaseBorderColor|;
        border-top:1px solid |phaseBorderColor|;
        border-bottom:1px solid |phaseBorderColor| }
      
      .phase a { color:|phaseColor| }

      .phaseC { margin:0 0 20px 0; border-left:1px solid |phaseBorderColor| }
      .phaseC > div { height:|pageCHeight|; width:|pageCWidth|; font-size:9pt;
                      margin:|pageMargin|; padding:|pageCPadding| }

      .picto { width:|pictoSize|; height:|pictoSize|; margin:|pictoMargin| }
      .pictoT { position:|epictoPosition|; top:8px; right:8px; width:22px;
                padding: |epictoPadding| }
      .pictoC, .pictoB { width:|pictoCSize|; height:|pictoCSize|;
                         margin:|pictoCMargin| }
      .inverted { filter:invert(66%) }

      .currentPage { background-color:|phaseBgcColor| !important;
                     color:|phaseCcolor| }
      .currentPage img { filter:invert(66%) }
      .currentPage a, .currentPage div { color:|brightColor| }

      // Animation for the "edit" icon
      @keyframes shakeImage {
        0% { transform:rotate(0deg); filter:sepia(0) }
        95% { transform:rotate(0deg); filter:sepia(0) }
        97% { transform:rotate(10deg); filter:sepia(1) }
        99% { transform:rotate(-10deg); filter:sepia(0.6) }
        100% { transform:rotate(0deg); filter:sepia(0.3) }
      }
      .shake { animation:shakeImage 10s infinite }''')

    def __init__(self, name, phases):
        # The name of the phase
        self.name = name
        # Its translated label
        o = phases.o
        self.label = o.translate('%s_phase_%s' % (o.class_.name, name))
        # A link to the global UiPhases object
        self.container = phases
        # The included pages, as a dict of UiPage instances
        self.pages = collections.OrderedDict()
        # The pages that were already walked, but that must not be shown
        self.hidden = {}

    def addPage(self, page):
        '''Adds page-related information in the phase'''
        # If the page is already there, we have nothing more to do
        if (page.name in self.pages) or (page.name in self.hidden): return
        # Add the page only if it must be shown
        o = self.container.o
        currentPage = self.container.currentPage
        showOnView = page.isShowable(o, 'view')
        showOnEdit = page.isShowable(o, 'edit')
        if showOnView or showOnEdit:
            # The page must be added
            self.pages[page.name] = self.container.lastPage = \
              UiPage(self, page, showOnView, showOnEdit)
            # Set it as current page when relevant
            if currentPage is None and \
               (page.name == self.container.currentPageName):
                self.container.currentPage = self.container.lastPage
        else:
            # The page must be hidden, and we must remember that fact
            self.hidden[page.name] = None

    def showPages(self, singlePhase, isEdit):
        '''Pages for this phase must not be shown if there is a single,
           not-editable page within a single phase.'''
        # If we are on an "edit" layout, the "pages" zone must always be shown
        # because it will contain at least the "save" and "cancel" buttons.
        if isEdit: return True
        if singlePhase and (len(self.pages) == 1):
            for page in self.pages.values():
                return page.showOnEdit
        return True

    def __repr__(self):
        if self.pages:
            pages = ': page(s) %s' % ','.join(list(self.pages.keys()))
        else:
            pages = ' (empty)'
        return '<phase %s%s>' % (self.name, pages)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class UiPhases:
    '''Represents the (sub-)set of phases and pages for some object that the
       currently logged user may see or edit.'''

    # PX displaying all phases of a given object
    view = Px('''
     <x var="mayEdit=guard.mayEdit(o);
             singlePhase=phases.singlePhase();
             id=o.iid">

      <!-- Single phase: display its pages -->
      <x if="singlePhase or isEdit"
         var2="phase=phases.default">:phase.view</x>

      <!-- Several phases: add a tab for every phase -->
      <x if="not singlePhase and not isEdit">
       <table cellpadding="0" cellspacing="0">
        <!-- First row: the tabs -->
        <tr><td class="tabSep">
         <table class="tabs" id=":'tabs_%d' % id">
          <tr valign="middle">
           <x for="phase in phases.all.values()"
              var2="suffix='%d_%s' % (id, phase.name);
                    tabId='tab_%s' % suffix">
            <td id=":tabId" class="tab">
             <a onclick=":'showTab(%s)'% q(suffix)"
                class="clickable">:phase.label</a>
            </td>
           </x>
          </tr>
         </table>
        </td></tr>
        <!-- Other rows: the fields -->
        <tr for="phase in phases.all.values()"
            id=":'tabcontent_%d_%s' % (id, phase.name)"
          style=":(loop.phase.nb==0) and 'display:table-row' or 'display:none'">
         <td>:phase.view</td>
        </tr>
       </table>
       <script>:'initTab(%s,%s)' % \
             (q('tab_%d' % id), q('%d_%s' % (id, phases.default.name)))</script>
      </x>
      <script>:'initPages(%d)' % phases.getStickyIndex(table)</script>
     </x>''',

     js = '''
      function invertPage(div, on) {
        // Do nothing if p_div correspond to the currently selected page
        if (div.id == 'currentPage') return;
        // Switch the page to dark mode
        var classes = div.classList,
            inverted = classes.contains('currentPage');
        if (on && !inverted) classes.add('currentPage');
        else if (!on && inverted) classes.remove('currentPage');
      }

      function initPages(rowIndex) {
        // Inject code allowing to switch between the (un)selected page view
        var phases = document.getElementsByClassName('phase'),
            divs=null, div=null;
        for (var i=0; i < phases.length; i++) {
          divs = phases[i].childNodes;
          for (var j=0; j < divs.length; j++) {
            div = divs[j];
            if (div.tagName == 'DIV') {
              div.addEventListener('mouseenter',
                                   function(){invertPage(this,true)});
              div.addEventListener('mouseleave',
                                   function(){invertPage(this,false)});
            }
          }
        }
        if (rowIndex != -1) {
          // Make the controls sticky
          var table = document.getElementById('pageLayout'),
              firstCell = table.childNodes[0].rows[rowIndex].cells[0];
          firstCell.classList.add('sticky');
        }
      }''')

    def __init__(self, page, o, layout):
        '''Initialises a structure allowing to render currently visible phases
           for p_o, for which the page named p_page is currently shown.'''
        # The object for which phases must be produced
        self.o = o
        # The layout
        self.layout = layout
        # The involved phases will be stored here, as UiPhase instances
        self.all = collections.OrderedDict()
        # The default phase (= the first encountered one)
        self.default = None
        # The currently shown page
        self.currentPageName = page
        self.currentPage = None
        # p_o's default page on p_layout
        self.defaultPageName = o.getDefaultPage(layout)
        # The last inserted page
        self.lastPage = None

    def singlePhase(self):
        '''Returns True if there is a single phase'''
        return len(self.all) == 1

    def singlePage(self):
        '''Returns True if there is a single page within a single phase'''
        if not self.singlePhase(): return
        return len(self.default.pages) == 1

    def addField(self, field):
        '''A new p_field has been encountered: it implies updating phases and
           pages.'''
        # Insert p_fields' phase and page into p_self.all
        phase = field.page.phase
        if phase not in self.all:
            uiPhase = self.all[phase] = UiPhase(phase, self)
            # Set it as default phase if there is no default phase yet
            if self.default is None:
                self.default = uiPhase
        else:
            uiPhase = self.all[phase]
        uiPhase.addPage(field.page)

    def unshowableCurrentPage(self):
        '''Return True if the current page can't be shown'''
        # The default page is always supposed to be showable
        if self.currentPageName == self.defaultPageName: return
        current = self.currentPage
        return not current or not current.showable(self.layout)

    def finalize(self):
        '''Finalize phases, ie, by removing phases without any visible page'''
        # Remove phases without page
        toDelete = None
        for name in self.all.keys():
            if not self.all[name].pages:
                if toDelete is None: toDelete = []
                toDelete.append(name)
        if toDelete is not None:
            for name in toDelete:
                del(self.all[name])
        if not self.currentPage:
            # We can't continue. Possibly, the page mentioned in the URL does
            # not exist, or the user may not be allowed to consult it anymore.
            # Typically, the user has logged out in browser tab A, and he tries
            # to reload a page in tab B, that became invisible when consulted as
            # anonymous user.
            # ~~~
            # Another explanation may be that there is no current page to show
            # according to conditions defined on pages for self.o's class. While
            # this can be a problem of conception in the app, there is a
            # possible solution to overcome it: defining, on this class, methods
            # m_getDefaultViewPage and / or m_getDefaultEditPage.
            o = self.o
            o.raiseMessage(o.translate('page_not_found'))

    def getStickyIndex(self, table):
        '''Returns the index of the element, within the page layout represented
           by p_table, that must be made sticky.'''
        # If stickyness must not be enabled for the current page, return -1
        sticky = self.currentPage.page.sticky
        sticky = (sticky == self.layout) if isinstance(sticky, str) else sticky
        return table.getRowIndex('b') if sticky else -1
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
from appy import Object
from appy.gen import utils as gutils

# ------------------------------------------------------------------------------
class Page:
    '''Used for describing a page, its related phase, show condition, etc.'''
    subElements = ('save', 'cancel', 'previous', 'next', 'edit')

    def __init__(self, name, phase='main', show=True, showSave=True,
                 showCancel=True, showPrevious=False, showNext=False,
                 showEdit=True, label=None):
        self.name = name
        self.phase = phase
        self.show = show
        # When editing the page, must I show the "save" button?
        self.showSave = showSave
        # When editing the page, must I show the "cancel" button?
        self.showCancel = showCancel
        # When editing the page, and when a previous page exists, must I show
        # the "previous" button?
        self.showPrevious = showPrevious
        # When editing the page, and when a next page exists, must I show the
        # "next" button?
        self.showNext = showNext
        # When viewing the page, must I show the "edit" button?
        self.showEdit = showEdit
        # Instead of computing a translated label, one may give p_label, a
        # fixed label which will not be translated.
        self.label = label

    @staticmethod
    def get(pageData):
        '''Produces a Page instance from p_pageData. User-defined p_pageData
           can be:
           (a) a string containing the name of the page;
           (b) a string containing <pageName>_<phaseName>;
           (c) a Page instance.
           This method returns always a Page instance.'''
        res = pageData
        if res and isinstance(res, basestring):
            # Page data is given as a string.
            pageElems = pageData.rsplit('_', 1)
            if len(pageElems) == 1: # We have case (a)
                res = Page(pageData)
            else: # We have case (b)
                res = Page(pageData[0], phase=pageData[1])
        return res

    def isShowable(self, obj, layoutType, elem='page'):
        '''Is this page showable for p_obj on p_layoutType ("view" or "edit")?

           If p_elem is not "page", this method returns the fact that a
           sub-element is viewable or not (buttons "save", "cancel", etc).'''
        # Define what attribute to test for "showability"
        attr = (elem == 'page') and 'show' or ('show%s' % elem.capitalize())
        # Get the value of the "show" attribute as identified above
        res = getattr(self, attr)
        if callable(res):
            res = gutils.callMethod(obj.appy(), res)
        if isinstance(res, str): return res == layoutType
        return res

    def getInfo(self, obj, layoutType):
        '''Gets information about this page, for p_obj, as an object.'''
        res = Object()
        for elem in Page.subElements:
            showable = self.isShowable(obj, layoutType, elem)
            setattr(res, 'show%s' % elem.capitalize(), showable)
        return res

    def getLabel(self, zobj):
        '''Returns the i18n label for this page, or a fixed label if self.label
           is not empty.'''
        label = self.label or '%s_page_%s' % (zobj.meta_type, self.name)
        return zobj.translate(label)
# ------------------------------------------------------------------------------



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import re, sys
from xml.sax.saxutils import quoteattr

from appy.pod import PodError
from appy.pod.elements import *
from appy.pod import actions, getUuid
from appy.shared.utils import Traceback
from appy.shared.xml_parser import xmlPrologue, Escape

# ------------------------------------------------------------------------------
class ParsingError(Exception): pass

# ParsingError-related constants -----------------------------------------------
ELEMENT = 'identifies the part of the document that will be impacted ' \
  'by the command. It must be one of %s.' % str(PodElement.POD_ELEMS)
FOR_EXPRESSION = 'must be of the form: {name} in {expression}. {name} must be '\
  'a Python variable name. It is the name of the iteration variable. ' \
  '{expression} is a Python expression that, when evaluated, produces a ' \
  'Python sequence (tuple, string, list, etc).'
POD_STATEMENT = 'A Pod statement has the form: do {element} ' \
  '[{command} {expression}]. {element} ' + ELEMENT + ' Optional {command} ' \
  'can be "if" (conditional inclusion of the element) or "for" (multiple ' \
  'inclusion of the element). For an "if" command, {expression} is any ' \
  'Python expression. For a "for" command, {expression} '+ FOR_EXPRESSION
FROM_CLAUSE = 'A "from" clause has the form: from[+] {expression}, where ' \
  '{expression} is a Python expression that, when evaluated, produces a valid '\
  'chunk of odt content that will be inserted instead of the element that is ' \
  'the target of the note. If the "+" is specified, the root target element ' \
  'will be kept and its content will be replaced with the expression result.'
EMPTY_NOTE = 'This note is empty. It must at least contain a Pod statement. '+ \
  POD_STATEMENT
BAD_STATEMENT = 'Syntax error for statement "%s". ' + POD_STATEMENT
BAD_SUB_STATEMENT = 'Wrong sub-statement "%s".'
BAD_ELEMENT = 'Bad element "%s". An element ' + ELEMENT
BAD_MINUS = "The '-' operator can't be used with element '%s'. It can only be "\
  "specified for elements among %s."
ELEMENT_NOT_FOUND = 'Action specified element "%s" but available elements ' \
  'in this part of the document are %s.'
BAD_FROM_CLAUSE = 'Syntax error in "from" clause "%s". ' + FROM_CLAUSE
DUPLICATE_NAMED_IF = 'An "if" statement with the same name already exists.'
ELSE_NOT_MAIN = 'An "else" clause can only be defined as a main statement.'
ELSE_WITHOUT_IF = 'No previous "if" statement could be found for this "else" ' \
  'statement.'
ELSE_WITHOUT_NAMED_IF = 'I could not find an "if" statement named "%s".'
BAD_FOR_EXPRESSION = 'Bad "for" expression "%s". A "for" expression ' + \
  FOR_EXPRESSION
BAD_VAR_EXPRESSION = 'Bad variable definition "%s". A variable definition ' \
  'must have the form {name[,name2,...]} = {expression}. Every name must be ' \
  'a Python-compliant variable name. {expression} is a Python expression. When ' \
  'encountering such a statement, pod will define, in the specified part of ' \
  'the document, variables {name}, {name2}, etc, whose values will be ' \
  'initialized by evaluating {expression}.'
BAD_VAR_NAME = 'Bad variable name "%s".'
EVAL_EXPR_ERROR = 'Error while evaluating expression "%s". %s'
BAD_META_CONDITION = 'Wrong meta-condition "%s". A meta-condition must be a ' \
  'Python expression surrounded by single or double quotes.'

# ------------------------------------------------------------------------------
class BufferIterator:
    def __init__(self, buffer):
        self.buffer = buffer
        self.remainingSubBufferIndexes = buffer.subBuffers.keys()
        self.remainingElemIndexes = buffer.elements.keys()
        self.remainingSubBufferIndexes.sort()
        self.remainingElemIndexes.sort()

    def __iter__(self): return self
    def __next__(self):
        # Stop if there is no more elem nor sub-buffer
        if not self.remainingSubBufferIndexes and not self.remainingElemIndexes:
            raise StopIteration
        nextSubBufferIndex = None
        if self.remainingSubBufferIndexes:
            nextSubBufferIndex = self.remainingSubBufferIndexes[0]
        nextExprIndex = None
        if self.remainingElemIndexes:
            nextExprIndex = self.remainingElemIndexes[0]
        # Compute min between nextSubBufferIndex and nextExprIndex
        if (nextSubBufferIndex != None) and (nextExprIndex != None):
            res = min(nextSubBufferIndex, nextExprIndex)
        elif (nextSubBufferIndex == None) and (nextExprIndex != None):
            res = nextExprIndex
        elif (nextSubBufferIndex != None) and (nextExprIndex == None):
            res = nextSubBufferIndex
        # Update "remaining" lists
        if res == nextSubBufferIndex:
            self.remainingSubBufferIndexes = self.remainingSubBufferIndexes[1:]
            resDict = self.buffer.subBuffers
        elif res == nextExprIndex:
            self.remainingElemIndexes = self.remainingElemIndexes[1:]
            resDict = self.buffer.elements
        return res, resDict[res]
    next = __next__ # Python2-3 compliance

# ------------------------------------------------------------------------------
class Buffer:
    '''Abstract class representing any buffer used during rendering'''
    elementRex = re.compile('([\w-]+:[\w-]+)\s*(.*?)>', re.S)

    def __init__(self, env, parent):
        self.parent = parent
        self.subBuffers = {} # ~{i_bufferIndex: Buffer}~
        self.env = env
        # Are we computing for pod (True) or px (False)
        self.pod = env.__class__.__name__ != 'PxEnvironment'

    def addSubBuffer(self, subBuffer=None):
        if not subBuffer:
            subBuffer = MemoryBuffer(self.env, self)
        self.subBuffers[self.getLength()] = subBuffer
        subBuffer.parent = self
        return subBuffer

    def removeLastSubBuffer(self):
        subBufferIndexes = self.subBuffers.keys()
        subBufferIndexes.sort()
        lastIndex = subBufferIndexes.pop()
        del self.subBuffers[lastIndex]

    def write(self, something): pass # To be overridden

    def getLength(self): pass # To be overridden

    def patchProtected(self, elem, attrs):
        '''Add or patch the attribute allowing to protect this p_elem (a cell or
           section), by evaluating variable "cellProtected" or
           "sectionProtected".'''
        type = elem[:elem.index(':')]
        name = '%s:protected' % type
        var = (type == 'table') and 'cell' or 'section'
        value = attrs.get(name, "false").strip() or "false"
        attrs[name] = ':str(%sProtected).lower()|"%s"' % (var, value)

    def patchElement(self, elem, attrs):
        '''Modifies p_elem's attributes when relevant and make them dependent of
           some variable definitions.'''
        if elem == 'table:table':
            # If a variable "tableName" is defined in the context, the current
            # table will get this name instead of its predefined name. Useful,
            # ie, for naming calc sheets (which are implemented as tables).
            attrs = attrs._attrs
            name = 'table:name'
            currentName = attrs.get(name) or getUuid(removeDots=True,prefix='T')
            attrs[name] = ':tableName|"%s"' % currentName
        elif elem == 'table:table-column':
            # Convert attribute "number-columns-repeated" of every table column
            # (or add it if it does not exist) to let the user define how he
            # will repeat table columns via variable "columnsRepeated".
            attrs = attrs._attrs
            key = 'table:number-columns-repeated'
            columnNumber = self.env.getTable().nbOfColumns -1
            nb = (key in attrs) and attrs[key] or '1'
            attrs[key] = ':columnsRepeated[%d]|%s' % (columnNumber, nb)
        elif elem == 'table:table-cell':
            attrs = attrs._attrs
            key = 'table:style-name'
            style = attrs.get(key)
            if style and style.startswith('S'):
                # By convention, if a style name starts with letter "S", POD
                # will allow to replace the name of the cell style with another
                # one, via function "styleFor" if found in the context. The
                # function will be passed 2 args: the name of the style that
                # should normally have been applied, and the context itself. It
                # must return the name of the style to apply. This mechanism is
                # restricted to styles starting with "S" for performance
                # reasons.
                attrs[key] = ':styleFor("%s",_ctx_)|"%s"' % (style, style)
            elif self.env.parser.caller.protection:
                self.patchProtected(elem, attrs)
        elif elem == 'text:section' and self.env.parser.caller.protection:
            self.patchProtected(elem, attrs._attrs)
        elif elem == 'text:list-item' and attrs.has_key('text:start-value'):
            attrs = attrs._attrs
            val = int(attrs['text:start-value'])
            attrs['text:start-value'] = ':loop._all_[-1].nb+%d|%d' % (val, val)

    def dumpStartElement(self, elem, attrs={}, ignoreAttrs=(), hook=False,
                         noEndTag=False, renamedAttrs=None):
        '''Inserts into this buffer the start tag p_elem, with its p_attrs,
           excepted those listed in p_ignoreAttrs. Attrs can be dumped with an
           alternative name if specified in dict p_renamedAttrs. If p_hook is
           not None (works only for MemoryBuffers), we will insert, at the end
           of the list of dumped attributes:
           * [pod] an Attributes instance, in order to be able, when evaluating
                   the buffer, to dump additional attributes, not known at this
                   dump time;
           * [px]  an Attribute instance, representing a special HTML attribute
                   like "checked" or "selected", that, if the tied expression
                   returns False, must not be dumped at all. In this case,
                   p_hook must be a tuple (s_attrName, s_expr).
        '''
        self.write('<%s' % elem)
        # Some table elements must be patched (pod only)
        if self.pod: self.patchElement(elem, attrs)
        for name, value in attrs.items():
            if ignoreAttrs and (name in ignoreAttrs): continue
            if renamedAttrs and (name in renamedAttrs): name=renamedAttrs[name]
            # If the value begins with ':', it is a Python expression. Else,
            # it is a static value.
            if not value.startswith(':'):
                self.write(' %s=%s' % (name, quoteattr(value)))
            else:
                self.write(' %s="' % name)
                self.addExpression(value[1:])
                self.write('"')
        res = None
        if hook:
            if self.pod:
                res = self.addAttributes()
            else:
                self.addAttribute(*hook)
        # Close the tag
        self.write(noEndTag and '/>' or '>')
        return res

    def dumpEndElement(self, elem):
        self.write('</%s>' % elem)

    def dumpElement(self, elem, content=None, attrs={}):
        '''For dumping a whole element at once'''
        self.dumpStartElement(elem, attrs)
        if content:
            self.dumpContent(content)
        self.dumpEndElement(elem)

    def dumpContent(self, content):
        '''Dumps string p_content into the buffer'''
        # For POD, take care of converting line breaks and tabs
        if self.pod:
            flavour = self.env.parser.caller.tabbedCR and 'odf*' or 'odf'
        else:
            flavour = 'xml'
        content = Escape.xml(content, flavour=flavour)
        self.write(content)

# ------------------------------------------------------------------------------
class FileBuffer(Buffer):
    def __init__(self, env, result):
        Buffer.__init__(self, env, None)
        self.result = result
        self.content = file(result, 'w')
        self.content.write(xmlPrologue)

    # getLength is used to manage insertions into sub-buffers. But in the case
    # of a FileBuffer, we will only have 1 sub-buffer at a time, and we don't
    # care about where it will be inserted into the FileBuffer.
    def getLength(self): return 0

    def write(self, something):
        try:
            self.content.write(something.encode('utf-8'))
        except UnicodeDecodeError:
            self.content.write(something)

    def pushSubBuffer(self, subBuffer): pass
    def getRootBuffer(self): return self

    def addExpression(self, expression, elem=None, tiedHook=None):
        try:
            expr = Expression(expression, self.pod)
            if tiedHook: tiedHook.tiedExpression = expr
            res, escape = expr.evaluate(self.env.context)
            if escape: self.dumpContent(res)
            else: self.write(res)
        except Exception, e:
            if not self.env.raiseOnError:
                PodError.dump(self, EVAL_EXPR_ERROR % (expression, e),
                              dumpTb=False)
            else:
                raise Exception(EVAL_EXPR_ERROR % (expression, e))

    def addAttributes(self):
        # Into a FileBuffer, it is not possible to insert Attributes. Every
        # Attributes instance is tied to an Expression; because dumping
        # expressions directly into FileBuffer instances seems to be rare, it
        # should not be a severe problem.
        pass

# ------------------------------------------------------------------------------
class MemoryBuffer(Buffer):
    class Rex:
        '''Regular expressions in use for a memory buffer'''
        part = '(for|if|else|with|meta-if)\s*(.*)'
        action = re.compile('(?:(\w+)\s*\:\s*)?do\s+(\w+)(-)?(?:\s+%s)?' % part)
        subAction = re.compile(part)
        for_ = re.compile('\s*([\w\-_,\s]+)\s+in\s+(.*)')
        vars_ = re.compile('\s*([\w\-_,\s@]+)\s*=\s*(.*)')
        var_ = re.compile('@?[\w\-_]+')
        from_ = re.compile('from(\+)?\s+(.*)')

    def __init__(self, env, parent):
        Buffer.__init__(self, env, parent)
        self.content = u''
        self.elements = {}
        self.action = None

    def clone(self):
        '''Produces an empty buffer that is a clone of this one'''
        return MemoryBuffer(self.env, self.parent)

    def addSubBuffer(self, subBuffer=None):
        sub = Buffer.addSubBuffer(self, subBuffer)
        # Dump a whitespace to avoid having several subbuffers referenced at the
        # same place within this buffer.
        self.content += ' '
        return sub

    def getInsertIndex(self, tag, opening):
        '''Return, within this buffer or any of its sub-buffers, the index that
           immediately follows the first occurrence of p_tag, which is an
           opening or ending tag, depending on p_opening.'''
        # More precisely, it returns a tuple (buffer, i), "buffer" being the
        # precise buffer where p_tag is found (can be p_self or one of its
        # sub-buffers at any level).
        buf = None
        i = self.content.find(tag)
        if i != -1:
            # The tag has been found directly in p_self
            if opening:
                # Find the end of the opening tag
                i = self.content.index('>', i) + 1
            else:
                # Get the index of the char following the closing tag
                i += len(tag)
            buf = self
        if self.subBuffers:
            # Browse sub-buffers
            subs = self.subBuffers.items()
            subs.sort()
            for j, sub in subs:
                buffer, k = sub.getInsertIndex(tag, opening)
                if buffer is None: continue
                # We have found the tag in this buffer
                if (buf is None) or (k < i):
                    # If the tag was also found in the main buffer, the tag
                    # from the sub-buffer must be chose because it comes first.
                    buf = buffer
                    i = k
                    break
        return buf, i

    def insertSubBuffer(self, subBuffer, after=None, afterClosing=None):
        '''Inserts p_subBuffer just after the main opening tag, in p_self that
           is already complete.'''
        # If p_after is given, it is a tag, and the p_subBuffer is inserted
        # immediately after this tag's start tag. If p_afterClosing is given, it
        # is a tag, and the p_subBuffer is inserted immediately after this tags'
        # closing tag. If p_both p_afterClosing and p_after are specified, the
        # method tries first to insert the subBuffer via p_afterClosing, and, if
        # this tag is not found, tries to insert it via p_after.
        buf = i = None
        if afterClosing:
            buf, i = self.getInsertIndex('</%s>' % afterClosing, opening=False)
        if after and (buf is None):
            buf, i = self.getInsertIndex('<%s' % after, opening=True)
        if buf is None:
            # Insert p_subBuffer after the main element
            buf = self
            i = buf.content.index('>') + 1
        # Insert sub-buffer's content info buf.content
        size = len(subBuffer.content)
        buf.content = buf.content[:i] + subBuffer.content + buf.content[i:]
        # Increment by 1 all subsequent elements and sub-buffers
        for subType in ('elements', 'subBuffers'):
            subElements = getattr(buf, subType)
            if not subElements: continue
            indexes = subElements.keys()
            indexes.sort(reverse=True)
            for j in indexes:
                if j >= i:
                    elem = subElements[j]
                    del subElements[j]
                    subElements[j+size] = elem

    def getRootBuffer(self):
        '''Returns the root buffer. For POD it is always a FileBuffer. For PX,
           it is a MemoryBuffer.'''
        if self.parent: return self.parent.getRootBuffer()
        return self

    def getLength(self): return len(self.content)

    def write(self, thing): self.content += thing

    def getIndex(self, podElemName):
        res = -1
        for index, podElem in self.elements.iteritems():
            if podElem.__class__.__name__.lower() == podElemName:
                if index > res:
                    res = index
        return res

    def getMainElement(self):
        res = None
        if self.elements.has_key(0):
            res = self.elements[0]
        return res

    def isMainElement(self, elem):
        '''Is p_elem the main element within this buffer?'''
        mainElem = self.getMainElement()
        if not mainElem: return
        if hasattr(mainElem, 'OD'): mainElem = mainElem.OD.elem
        if elem != mainElem: return
        # elem is the same as the main elem. But is it really the main elem, or
        # the same elem, found deeper in the buffer?
        for index, iElem in self.elements.iteritems():
            foundElem = None
            if hasattr(iElem, 'OD'):
                if iElem.OD:
                    foundElem = iElem.OD.elem
            else:
                foundElem = iElem
            if (foundElem == mainElem) and (index != 0):
                return
        return True

    def unreferenceElement(self, elem):
        # Find last occurrence of this element
        elemIndex = -1
        for index, iElem in self.elements.iteritems():
            foundElem = None
            if hasattr(iElem, 'OD'):
                # A POD element
                if iElem.OD:
                    foundElem = iElem.OD.elem
            else:
                # A PX elem
                foundElem = iElem
            if (foundElem == elem) and (index > elemIndex):
                elemIndex = index
        del self.elements[elemIndex]

    def pushSubBuffer(self, subBuffer):
        '''Sets p_subBuffer at the very end of the buffer.'''
        subIndex = None
        for index, aSubBuffer in self.subBuffers.iteritems():
            if aSubBuffer == subBuffer:
                subIndex = index
                break
        if subIndex != None:
            # Indeed, it is possible that this buffer is not referenced
            # in the parent (if it is a temp buffer generated from a cut)
            del self.subBuffers[subIndex]
            self.subBuffers[self.getLength()] = subBuffer
            self.content += u' '

    def transferAllContent(self):
        '''Transfer all content to parent'''
        if isinstance(self.parent, FileBuffer):
            # First unreference all elements
            for index in self.getElementIndexes(expressions=False):
                del self.elements[index]
            self.evaluate(self.parent, self.env.context)
        else:
            # Transfer content in itself
            oldParentLength = self.parent.getLength()
            self.parent.write(self.content)
            # Transfer elements
            for index, podElem in self.elements.iteritems():
                self.parent.elements[oldParentLength + index] = podElem
            # Transfer sub-buffers
            for index, buf in self.subBuffers.iteritems():
                self.parent.subBuffers[oldParentLength + index] = buf
        # Empty the buffer
        MemoryBuffer.__init__(self, self.env, self.parent)
        # Change buffer position wrt parent
        self.parent.pushSubBuffer(self)

    def addElement(self, elem, elemType='pod'):
        if elemType == 'pod':
            elem = PodElement.create(elem)
        self.elements[self.getLength()] = elem
        if isinstance(elem, Cell) or isinstance(elem, Table):
            elem.tableInfo = self.env.getTable()
            if isinstance(elem, Cell):
                # Remember where this cell is in the table
                elem.colIndex = elem.tableInfo.curColIndex
        if elem == 'x':
            # See comment on similar statement in the method below.
            self.content += u' '

    def addExpression(self, expression, elem=None, tiedHook=None):
        '''Creates an Expression instance and add it in the buffer'''
        # Create the POD expression
        expr = Expression(expression, self.pod)
        # Get the meta-condition if found
        if elem and elem.attrs:
            metaCondition = elem.attrs.get('text:condition')
            if metaCondition and (metaCondition.lower() != 'ooow:true'):
                metaCondition = metaCondition.split(':', 1)[1]
                metaWrap = metaCondition[0]
                if (metaWrap != metaCondition[-1]) or \
                   (metaWrap not in expr.metaWraps):
                    raise ParsingError(BAD_META_CONDITION % metaCondition)
                expr.metaWrap = metaWrap
                expr.metaCondition = metaCondition.strip('"\'"')
        if tiedHook: tiedHook.tiedExpression = expr
        self.elements[self.getLength()] = expr
        # To be sure that an expr and an elem can't be found at the same index
        # in the buffer.
        self.content += u' '

    def addAttributes(self):
        '''pod-only: adds an Attributes instance into this buffer'''
        attrs = Attributes(self.env)
        self.elements[self.getLength()] = attrs
        self.content += u' '
        return attrs

    def addAttribute(self, name, expr):
        '''px-only: adds an Attribute instance into this buffer'''
        attr = Attribute(name, expr)
        self.elements[self.getLength()] = attr
        self.content += u' '
        return attr

    def _getVariables(self, expr):
        '''Returns variable definitions in p_expr as a list
           ~[(s_varName|[s_varName], s_expr)]~.'''
        r = []
        for sub in expr.strip().split(';'):
            # If the last variable definition ends with a semi-colon, it
            # produces an empty element.
            if not sub: continue
            # Do we have valid variable definitions ?
            match = self.Rex.vars_.match(sub)
            if not match:
                raise ParsingError(BAD_VAR_EXPRESSION % sub)
            # Extract individual variable definitions
            names = match.group(1).strip()
            if ',' in names:
                # Several variable names
                names = names.split(',')
                # Strip these names and ensure they are valid variable names
                i = 0
                while i < len(names):
                    name = names[i].strip()
                    if not self.Rex.var_.match(name):
                        raise ParsingError(BAD_VAR_NAME % name)
                    names[i] = name
                    i += 1
            r.append((names, match.group(2).strip()))
        return r

    def _getForIterators(self, iters):
        '''Gets the variables used as iterators in a "for" statement, from
           comma-separated string p_iters.'''
        r = iters.split(',')
        i = len(r) - 1
        while i >= 0:
            r[i] = r[i].strip()
            i -= 1
        return r

    def createPodAction(self, actionType, statements, statementName, subExpr,
                        podElem, minus, main=True):
        '''Creates an Action instance, depending on p_actionType'''
        if actionType == 'if':
            r = actions.If(statementName, self, subExpr, podElem, minus)
            if main:
                self.env.ifActions.append(r)
                if r.name:
                    # We must register this action as a named action
                    if self.env.namedIfActions.has_key(r.name):
                        raise ParsingError(DUPLICATE_NAMED_IF)
                    self.env.namedIfActions[r.name] = r
        elif actionType == 'else':
            if not main: raise ParsingError(ELSE_NOT_MAIN)
            if not self.env.ifActions: raise ParsingError(ELSE_WITHOUT_IF)
            # Does the "else" action reference a named "if" action?
            ifReference = subExpr.strip()
            if ifReference:
                if not self.env.namedIfActions.has_key(ifReference):
                    raise ParsingError(ELSE_WITHOUT_NAMED_IF % ifReference)
                linkedIfAction = self.env.namedIfActions[ifReference]
                # This "else" action "consumes" the "if" action: this way,
                # it is not possible to define two "else" actions related to
                # the same "if".
                del self.env.namedIfActions[ifReference]
                self.env.ifActions.remove(linkedIfAction)
            else:
                linkedIfAction = self.env.ifActions.pop()
            r = actions.Else(statementName, self, None, podElem, minus,
                             linkedIfAction)
        elif actionType == 'for':
            forRes = self.Rex.for_.match(subExpr.strip())
            if not forRes:
                raise ParsingError(BAD_FOR_EXPRESSION % subExpr)
            iters, subExpr = forRes.groups()
            iters = self._getForIterators(iters)
            r = actions.For(statementName, self, subExpr, podElem, minus, iters)
        elif actionType == 'with':
            variables = self._getVariables(subExpr)
            r = actions.Variables(statementName, self, podElem, minus,variables)
        elif actionType == 'meta-if':
            r = actions.MetaIf(self, subExpr, podElem, minus, statements)
        else:
            r = actions.Null(self, podElem)
        return r

    def createPodActions(self, statements):
        '''Tries to create action(s) based on p_statements. If the statement is
           not correct, r_ is -1. Else, r_ is the index of the element within
           the buffer that is the object of the action(s).'''
        r = -1
        try:
            # Check that the statement group is not empty
            if not statements: raise ParsingError(EMPTY_NOTE)
            # Get the main statement (starting with "do...") and check it
            main = statements[0]
            aRes = self.Rex.action.match(main)
            if not aRes:
                raise ParsingError(BAD_STATEMENT % main)
            statementName, podElem, minus, actionType, subExpr = aRes.groups()
            if not (podElem in PodElement.POD_ELEMS):
                raise ParsingError(BAD_ELEMENT % podElem)
            if minus and (not podElem in PodElement.MINUS_ELEMS):
                raise ParsingError(BAD_MINUS % (podElem,PodElement.MINUS_ELEMS))
            # Find the target element in the buffer
            i = self.getIndex(podElem)
            if i == -1:
                raise ParsingError(ELEMENT_NOT_FOUND % (podElem, str([
                        e.__class__.__name__.lower() \
                        for e in self.elements.values()])))
            podElem = self.elements[i]
            # Create the main action
            self.action = self.createPodAction(actionType, statements,
              statementName, subExpr, podElem, minus)
            # Parse the remaining statements, that can contain any number of
            # secondary actions and a from clause.
            fromClause = last = None
            for statement in statements[1:]:
                # Get the "from" clause
                if statement.startswith('from') or \
                   statement.startswith('from+'):
                    fromInfo = self.Rex.from_.match(statement)
                    if not fromInfo:
                        raise ParsingError(BAD_FROM_CLAUSE % fromClause)
                    fromClause = fromInfo.groups()
                # Get any secondary statement
                else:
                    info = self.Rex.subAction.match(statement)
                    if not info:
                        raise ParsingError(BAD_SUB_STATEMENT % statement)
                    actionType, subExpr = info.groups()
                    last = self.createPodAction(actionType, statements, '',
                                             subExpr, podElem, None, main=False)
                    self.action.addSubAction(last)
            # Link the "from" clause
            if fromClause:
                target = last or self.action
                target.setFrom(*fromClause)
            success, msg = self.action.check()
            if not success: raise ParsingError(msg)
            r = i
        except ParsingError, ppe:
            PodError.dump(self, str(ppe), removeFirstLine=True)
        return r

    def createPxAction(self, elem, actionType, statement):
        '''Creates a PX action and link it to this buffer. If an action is
           already linked to this buffer (in self.action), this action is
           chained behind the last action via self.action.subAction.'''
        res = 0
        statement = statement.strip()
        if actionType == 'for':
            forRes = self.Rex.for_.match(statement)
            if not forRes:
                raise ParsingError(BAD_FOR_EXPRESSION % statement)
            iters, subExpr = forRes.groups()
            iters = self._getForIterators(iters)
            action = actions.For('for', self, subExpr, elem, False, iters)
        elif actionType == 'if':
            action = actions.If('if', self, statement, elem, False)
        elif actionType in ('var', 'var2'):
            variables = self._getVariables(statement)
            action = actions.Variables('var', self, elem, False, variables)
        # Is it the first action for this buffer or not?
        if not self.action:
            self.action = action
        else:
            self.action.addSubAction(action)
        return res

    def cut(self, index, keepFirstPart):
        '''Cuts this buffer into 2 parts. Depending on p_keepFirstPart, the 1st
        (from 0 to index-1) or the second (from index to the end) part of the
        buffer is returned as a MemoryBuffer instance without parent; the other
        part is self.'''
        res = MemoryBuffer(self.env, None)
        # Manage buffer meta-info (elements, expressions, subbuffers)
        subBuffersToDelete = []
        elementsToDelete = []
        mustShift = False
        for itemIndex, item in BufferIterator(self):
            if keepFirstPart:
                if itemIndex >= index:
                    newIndex = itemIndex-index
                    if isinstance(item, MemoryBuffer):
                        res.subBuffers[newIndex] = item
                        subBuffersToDelete.append(itemIndex)
                    else:
                        res.elements[newIndex] = item
                        elementsToDelete.append(itemIndex)
            else:
                if itemIndex < index:
                    if isinstance(item, MemoryBuffer):
                        res.subBuffers[itemIndex] = item
                        subBuffersToDelete.append(itemIndex)
                    else:
                        res.elements[itemIndex] = item
                        elementsToDelete.append(itemIndex)
                else:
                    mustShift = True
        if elementsToDelete:
            for elemIndex in elementsToDelete:
                del self.elements[elemIndex]
        if subBuffersToDelete:
            for subIndex in subBuffersToDelete:
                del self.subBuffers[subIndex]
        if mustShift:
            elements = {}
            for elemIndex, elem in self.elements.iteritems():
                elements[elemIndex-index] = elem
            self.elements = elements
            subBuffers = {}
            for subIndex, buf in self.subBuffers.iteritems():
                subBuffers[subIndex-index] = buf
            self.subBuffers = subBuffers
        # Manage content
        if keepFirstPart:
            res.write(self.content[index:])
            self.content = self.content[:index]
        else:
            res.write(self.content[:index])
            self.content = self.content[index:]
        return res

    def getElementIndexes(self, expressions=True):
        res = []
        for index, elem in self.elements.iteritems():
            condition = isinstance(elem, Expression) or \
                        isinstance(elem, Attributes)
            if not expressions:
                condition = not condition
            if condition:
                res.append(index)
        return res

    def transferActionIndependentContent(self, actionElemIndex):
        # Manage content to transfer to parent buffer
        if actionElemIndex != 0:
            actionIndependentBuffer = self.cut(actionElemIndex,
                                               keepFirstPart=False)
            actionIndependentBuffer.parent = self.parent
            actionIndependentBuffer.transferAllContent()
            self.parent.pushSubBuffer(self)
        # Manage content to transfer to a child buffer
        actionElemIndex = self.getIndex(
            self.action.elem.__class__.__name__.lower())
        # We recompute actionElemIndex because after cut it may have changed
        elemIndexes = self.getElementIndexes(expressions=False)
        elemIndexes.sort()
        if elemIndexes.index(actionElemIndex) != (len(elemIndexes)-1):
            # I must create a sub-buffer with the impactable elements after
            # the action-related element
            childBuffer = self.cut(elemIndexes[elemIndexes.index(
                actionElemIndex)+1], keepFirstPart=True)
            self.addSubBuffer(childBuffer)
            res = childBuffer
        else:
            res = self
        return res

    def getStartIndex(self, removeMainElems):
        '''When I must dump the buffer, sometimes (if p_removeMainElems is
        True), I must dump only a subset of it. This method returns the start
        index of the buffer part I must dump.'''
        if not removeMainElems: return 0
        # Find the start position of the deepest element to remove
        deepestElem = self.action.elem.DEEPEST_TO_REMOVE
        pos = self.content.find('<%s' % deepestElem.elem)
        pos = pos + len(deepestElem.elem)
        # Now we must find the position of the end of this start tag,
        # skipping potential attributes.
        inAttrValue = False # Are we parsing an attribute value ?
        endTagFound = False # Have we found the end of this tag ?
        while not endTagFound:
            pos += 1
            nextChar = self.content[pos]
            if (nextChar == '>') and not inAttrValue:
                # Yes we have it
                endTagFound = True
            elif nextChar == '"':
                inAttrValue = not inAttrValue
        return pos + 1

    def getStopIndex(self, removeMainElems):
        '''This method returns the stop index of the buffer part I must dump.'''
        if removeMainElems:
            ns = self.env.namespaces
            deepestElem = self.action.elem.DEEPEST_TO_REMOVE
            pos = self.content.rfind('</%s>' % deepestElem.getFullName(ns))
            res = pos
        else:
            res = self.getLength()
        return res

    def removeAutomaticExpressions(self):
        '''When a buffer has an action with minus=True, we must remove the
           expressions automatically inserted by pod ("columnsRepeat"
           "cellProtected"...) . Else, we will have problems when computing the
           index of the part to keep (m_getStartIndex).'''
        # Find the position of the end of the starting element of the deepest
        # element to remove.
        deepestElem = self.action.elem.DEEPEST_TO_REMOVE
        pos = self.content.find('<%s' % deepestElem.elem)
        pos = self.content.find('>', pos)
        for index in self.elements.keys():
            if index < pos: del self.elements[index]

    reTagContent = re.compile('<(?P<p>[\w-]+):(?P<f>[\w-]+)(.*?)>.*</(?P=p):' \
                              '(?P=f)>', re.S)
    def evaluate(self, result, context, subElements=True,
                 removeMainElems=False):
        '''Evaluates this buffer given the current p_context and add the result
           into p_result. With pod, p_result is the root file buffer; with px
           it is a memory buffer.'''
        if not subElements:
            # Dump the root tag in this buffer, but not its content
            res = self.reTagContent.match(self.content.strip())
            if not res: result.write(self.content)
            else:
                g = res.group
                g1, g2, g3 = g(1), g(2), g(3)
                if self.env.parser.caller.protection and (g2 == 'table-cell'):
                    # Attribute "protected" must be removed
                    g3 = g3.replace('table:protected=" "', '')
                r = '<%s:%s%s></%s:%s>' % (g1, g2, g3, g1, g2)
                result.write(r)
        else:
            if removeMainElems: self.removeAutomaticExpressions()
            currentIndex = self.getStartIndex(removeMainElems)
            for index, evalEntry in BufferIterator(self):
                result.write(self.content[currentIndex:index])
                currentIndex = index + 1
                if isinstance(evalEntry, Expression):
                    try:
                        res, escape = evalEntry.evaluate(context)
                        if escape: result.dumpContent(res)
                        else: result.write(res)
                    except actions.EvaluationError, e:
                        # This exception has already been treated (see the 
                        # "except" block below). Simply re-raise it when needed.
                        if self.env.raiseOnError: raise e
                    except Exception, e:
                        if not self.env.raiseOnError:
                            PodError.dump(result, EVAL_EXPR_ERROR % (
                                          evalEntry.expr, e))
                        else:
                            traceback = Traceback.get(5).decode('utf-8')
                            raise actions.EvaluationError(e, EVAL_EXPR_ERROR % \
                                             (evalEntry.expr, '\n' + traceback))
                elif isinstance(evalEntry, Attributes) or \
                     isinstance(evalEntry, Attribute):
                    result.write(evalEntry.evaluate(context))
                else: # It is a subBuffer
                    if evalEntry.action:
                        evalEntry.action.execute(result, context)
                    else:
                        result.write(evalEntry.content)
            stopIndex = self.getStopIndex(removeMainElems)
            if currentIndex < (stopIndex-1):
                result.write(self.content[currentIndex:stopIndex])

    def clean(self):
        '''Cleans the buffer content'''
        self.content = u''
# ------------------------------------------------------------------------------
